%  DIRECTIONALGRANULOMETRY Create object for granulometric analysis of images with directional SE's
%     g = GRANULOMETRY(TYPE,NORIENTATIONS) creates a directional granulometry 
%     of type TYPE based on bresenham structuring elements with NORIENTATIONS 
%     different orientations.
%  
%     g = GRANULOMETRY(TYPE,NORIENTATIONS, SCALES) creates a directional granulometry 
%     of type TYPE based on bresenham structuring elements with NORIENTATIONS 
%     different orientations for the scales determined by SCALES (see GRANULOMETRY)
%  
%     See also GRANULOMETRY
%
%
%       Please cite the following references:
%
%       [1] W. Liao, J. Chanussot, M. Dalla Mura, X. Huang, R. Bellens,
%           S. Gautama, W, Philips, "Taking Optimal Advantage of Fine Spatial 
%           Resolution: Promoting partial image reconstruction for the 
%           morphological analysis of very-high-resolution images",  IEEE 
%           Geoscience and Remote Sensing Magazine, vol. 5, no. 2, pp. 8-28, 2017.
%
%       [2] W. Liao, M. Dalla Mura, J. Chanussot, R. Bellens, W. Philips, 
%           "Morphological Attribute Profiles With Partial Reconstruction", 
%           IEEE Transactions on Geoscience and Remote Sensing, vol. 54, 
%           no. 3, pp. 1738-1756, 2016.
%
%       [3] R. Bellens, J. Chan, L. Martinez-Fonte, S. Gautama, F. Canters, 
%           W. Philips, "Improved classification of VHR images of urban areas 
%           using directional morphological profiles", IEEE Transactions on
%           Geoscience and Remote Sensing, vol. 46, no. 10, pp. 2803-2813, 2008.
%
%
%
%       Copyright notes
%       Author: Rik Bellens and Wenzhi Liao, IPI, Telin, Ghent University, Belgium
%       Date: 19/07/2016
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

classdef directionalgranulometry < granulometry
    
    properties
        norientations;
    end
        
    methods 
        function obj = directionalgranulometry(type,norientations,varargin)
            obj@granulometry(type,@(l,a)faststrel('bresenhamline',l,a),varargin{:});
            obj.norientations = norientations;
        end
        
        
    end

    methods (Access = public)
        function [out, R] = preoperation(obj,im,n)

            n = obj.scalegenerator(n);
            out = cast(-inf,class(im));
            R = 0;
            r = inf;

            for i=1:obj.norientations
                a = 180/obj.norientations*(i-1);
                l = floor(n*max(abs(sin(a/180*pi)),abs(cos(a/180*pi))));
                se = obj.segenerator(l,a);
                out = max(out,obj.openoperation(im,(se)));
                
                r = min(r,getradius(se));
                R = 0.1*max(R,getradius(se));%modified 2011-08-18  0.1*max(R,getradius(se));
            end
        end
    end
end
