%  GRANULOMETRY Create object for granulometric analysis of images with isotropic SE's
%     g = GRANULOMETRY('euclidean',SETYPE) creates a granulometry based on 
%     (euclidean) openings. 
%     
%     g = GRANULOMETRY('geodesic',SETYPE) creates a granulometry based on 
%     geodesic openings or openings with reconstruction.
%     
%     g = GRANULOMETRY('partialreconstruction',SETYPE) creates a granulometry
%     based on openings with partial reconstruction.
%     
%     g = GRANULOMETRY(OP,SETYPE) creates a granulometry based on the 
%     image operation defined by the function handle OP.
%     
%     In all the above, SETYPE defines the family of structuring elements. Valid values
%     are square, octagon, dodecagon, hexadecagon. SETYPE can also be a function handle 
%     generating structuring elements. 
%     
%     g = GRANULOMETRY(TYPE,SETYPE,SCALES) creates a granulometry of type TYPE with 
%     structuring elements of type SETYPE. SCALES is an array of increasing scales that 
%     is used to generate the structuring elements. 
%  
%     g = GRANULOMETRY(TYPE,SETYPE,SCALEFUNCTION) creates a granulometry of type TYPE with 
%     structuring elements of type SETYPE. SCALEFUNCTION is a handle to a function that 
%     generates the scales for the structuring elements. 
%  
%
%
%     Uses
%     ----
%     
%     p = GenerateFirstNScales(g,IM,N) generates the first N scales of the granulometry for 
%     the image IM. p is a 1-by-N cell array of the openings of IM.
%
%
%
%       Please cite the following references:
%
%       [1] W. Liao, J. Chanussot, M. Dalla Mura, X. Huang, R. Bellens,
%           S. Gautama, W, Philips, "Taking Optimal Advantage of Fine Spatial 
%           Resolution: Promoting partial image reconstruction for the 
%           morphological analysis of very-high-resolution images",  IEEE 
%           Geoscience and Remote Sensing Magazine, vol. 5, no. 2, pp. 8-28, 2017.
%
%       [2] W. Liao, M. Dalla Mura, J. Chanussot, R. Bellens, W. Philips, 
%           "Morphological Attribute Profiles With Partial Reconstruction", 
%           IEEE Transactions on Geoscience and Remote Sensing, vol. 54, 
%           no. 3, pp. 1738-1756, 2016.
%
%       [3] R. Bellens, J. Chan, L. Martinez-Fonte, S. Gautama, F. Canters, 
%           W. Philips, "Improved classification of VHR images of urban areas 
%           using directional morphological profiles", IEEE Transactions on
%           Geoscience and Remote Sensing, vol. 46, no. 10, pp. 2803-2813, 2008.
%
%
%
%       Copyright notes
%       Author: Wenzhi Liao and Rik Bellens, IPI, Telin, Ghent University, Belgium
%       Date: 19/07/2016
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
   

classdef granulometry < handle
    
    properties (Access = public)
        openoperation
        postoperation
        segenerator
        scalegenerator
    end
    
    methods
        
        function obj = granulometry(type,setype,scalegenerator)
            
            if isa(setype,'function_handle')
                obj.segenerator = setype;
            else
                switch setype
		    case {'disk'}%square
			obj.segenerator = @(i)faststrel('disk',i);
                    case {'line'}
                        obj.segenerator = @(i)faststrel('line',3*i+1,mod(90*i,360));           
                    case {'square'}
                        obj.segenerator = @(i)faststrel('square',2*i+1);                                    
                    case {'octagon'}
                        obj.segenerator = @(i)faststrel('octagon',i);
                    case {'dodecagon'}
                        obj.segenerator = @(i)faststrel('dodecagon',i);
                    case {'hexadecagon'}
                        obj.segenerator = @(i)faststrel('hexadecagon',i);
                    otherwise
                        error('..');
                end
            end
            
            if nargin<3
                obj.scalegenerator = @(i)i;
            elseif isa(scalegenerator,'functio_handle')
                obj.scalegenerator = scalegenerator;
            elseif isvector(scalegenerator)
                obj.scalegenerator = @(i)subsref(scalegenerator,substruct('()',{i}));
            end

            if isa(type,'function_handle')
                obj.openoperation = type;
            else
                switch type
                    case {'euclidean'}
                        obj.openoperation = @imopen;
                        obj.postoperation = @(im,oldim,scale,R)im;
                    case {'geodesic'}
                        obj.openoperation = @imerode;
                        obj.postoperation = @(im,oldim,scale,R)imreconstruct(im,oldim);
        		    case {'partialreconstruction'}
                        obj.openoperation = @imopen;
                        obj.postoperation = @(im,oldim,scale,R)partialreconstruct(im,oldim,ceil(R*(sqrt(2)-1)));
                    case {'complementabsorption'}
                        obj.openoperation = @imopen;
                        obj.postoperation = @(im,oldim,scale,R)granulometry.forceComplementAbsorption(im,oldim,@(i,s)preoperation(obj,i,s),scale);
                    otherwise
                        error('..');
                end
            end

	end
        
        function [out, R] = GenerateNthScale(obj,im,n)
            [out, R] = preoperation(obj,im,n);
            out = obj.postoperation(out,im,n,R);
        end
        
        function [out, R] = GenerateFirstNScales(obj,im,n)
            R = zeros(1,n);
            out = cell(1,n);
            for i=1:n
                tic;
                [out{i}, R(i)] = GenerateNthScale(obj,im,i);
		im = out{i};
                fprintf('scale %d (radius = %f): %f\n',i,R(i),toc);
            end
        end
        
	
	% experimental functions
        function [out,R] = GenerateContrastProfile(obj,im,n)
            R = zeros(1,n);
            out = cell(1,n);
	    start = 1;
	    
	    persistent cache;
	    if ~isempty(cache)
		if length(cache.out)>=n
		    out = cache.out(1:n);
		    R = cache.R(1:n);
		    return;
		else
		    start = length(cache.out)+1;
		    out(1:(start-1)) = cache.out;
		    R(1:(start-1)) = cache.R;
		end
		
	    end
		
            cim = imcomplement(im);
            for i=start:n
                tic;
                [open, R(i)] = GenerateNthScale(obj,im,i);
                [close, R(i)] = GenerateNthScale(obj,cim,i);
                out{i} = max(im-open,cim-close);
                fprintf('scale %d (radius = %f): %f\n',i,R(i),toc);
            end
	    cache.out = out;
	    cache.R = R;
         end
           
        
        function [m, dm] = SizeFromProfile(obj,im,n,t)
            [p,R] = GenerateContrastProfile(obj,im,n);
	    p = [zeros(size(p{end})),p];
	    p = cellfun(@double,p,'UniformOutput',false);
	    R = [0,R];
	    m = zeros(size(p{end}));
	    dm = zeros(size(p{end}));
	    for i=2:n
		m = m + max(0,min( (t-p{i-1})./(p{i}-p{i-1}) ,1))*(R(i)-R(i-1));
		I = (p{i-1}<t)&(p{i}>=t);
		dm(I) = (R(i)-R(i-1))./(p{i}(I)-p{i-1}(I));
	    end
        end
	
        function [m,t] = OptSizeFromProfile(obj,im)
	    t0 = 50;
	    n = 40;
	    t = fminsearch(@(t)SizeDerivative(obj,im,n,t),t0,optimset('TolX',1,'MaxIter',15));
	    m = SizeFromProfile(obj,im,n,t);
	end
	
	function d = SizeDerivative(obj,im,n,t)
	    [m,dm] = SizeFromProfile(obj,im,n,t);
	    d = mean(dm(m<max(m(:))));
	    fprintf('t = %2.2f, d = %f \n',t,d);
	end
        
	function m = SignedSizeFromProfile(obj,im,n,t)
	    
	    [p,R] = GenerateFirstNScales(obj,im,n);
	    p = [zeros(size(p{end}),class(p{1})),p];
	    p = cellfun(@(i)im-i,p,'UniformOutput',false);
	    p = cellfun(@double,p,'UniformOutput',false);
	    R = [0,R];
	    m = zeros(size(p{end}));
	    for i=2:n
		m = m + max(0,min( (t-p{i-1})./(p{i}-p{i-1}) ,1))*(R(i)-R(i-1));
	    end
	    m1 = m;
	    im = imcomplement(im);
	    [p,R] = GenerateFirstNScales(obj,im,n);
	    p = [zeros(size(p{end}),class(p{1})),p];
	    p = cellfun(@(i)im-i,p,'UniformOutput',false);
	    p = cellfun(@double,p,'UniformOutput',false);
	    R = [0,R];
	    m = zeros(size(p{end}));
	    for i=2:n
		m = m + max(0,min( (t-p{i-1})./(p{i}-p{i-1}) ,1))*(R(i)-R(i-1));
	    end
	    m = (m<m1).*m - (m>=m1).*m1;
	end
	
	
    end

    methods (Access = public)
        function [out, R] = preoperation(obj,im,n)
            i = obj.scalegenerator(n);
            se = obj.segenerator(i);
            out = obj.openoperation(im,se);
            R = getradius(se);
        end
    end
    
    methods (Static,Access = protected)
	% two-sided granulometry
	
	function out = forceComplementAbsorption(im,prevscale,openoperation,scale)
	    if scale==1
		out = im;
		return;
	    end
	    a = openoperation(prevscale-im,scale-1);
	    a = imreconstruct(a,prevscale-im);
	    out = prevscale-a;
	end
	
	
    end
end
